/* This file is part of Legendes II */
/* Copyright (c) 1998-1999 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File lcl.c */

/* Thus speaketh ye mighty CVS:
 * $Id: lcl.c,v 1.3 1999/05/30 19:57:05 david Exp $ */

/* This file contains the Legendes Client Library */

/* CVS identity of this file. */
static char id[] = "$Id: lcl.c,v 1.3 1999/05/30 19:57:05 david Exp $";

/* Include files */

#include <stdio.h>		/* Standard I/O */
#include <stdlib.h>		/* Standard library functions */
#include <string.h>

#include <unistd.h>		/* UNIX standard functions */

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include <netdb.h>		/* gethostbyname() and friends */

#include <errno.h>

#include <stdarg.h>		/* Variable number of arguments */


#include "lgdproto.h"

#include "lcl.h"

#include "lcl_p.h"

void
lcl_error (char *fmt, ...)
     /* Produce an error. */
     /* XXX - We would like to allow the user (i.e. the client) to
      * setup an error callback that this function would call.  Please
      * read the longjmp(2) man page. */
{
  va_list ap;

  fprintf (stderr, "lcl: ");
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  exit (1);
}

void
lcl_do_req (lcl_conn_t *conn, union lgd_proto_msg_u *req)
     /* Send a protocol request. */
{
  if ( write (conn->s, req, ntohl (req->f.len)) < 0 )
    lcl_error ("Can't send: %s\n", strerror (errno));
}

void
lcl_send_req (lcl_conn_t *conn, uint32_t code, uint32_t callback,
	      void *data, uint32_t datalen)
     /* Construct and send a protocol request.  The code argument is
      * passed in host order and this function converts it to network
      * order.  Of course, data is not byte-swapped in any way for
      * this function cannot how that should be done. */
{
  uint32_t len;
  union lgd_proto_msg_u *req;

  len = datalen+16;
  req = alloca (len);
  req->f.code = htonl (code);
  req->f.len = htonl (len);
  req->f.res = htonl (0);
  req->f.callback = callback; /* Keep host order */
  memcpy (req->f.data, data, datalen);
  lcl_do_req (conn, req);
}

void
lcl_req_0arg (lcl_conn_t *conn, uint32_t code, uint32_t callback)
     /* Construct and send a protocol request with 0 arguments.  Thus,
      * the code is the only important part. */
{
  lcl_send_req (conn, code, callback, NULL, 0);
}

void
lcl_req_1arg (lcl_conn_t *conn, uint32_t code, uint32_t callback,
	      uint32_t d)
     /* Construct and send a protocol request with 1 argument, namely
      * d.  d is passed in host order to this function and it converts
      * it to network order. */
{
  uint32_t hd;

  hd = htonl (d);
  lcl_send_req (conn, code, callback, &hd, sizeof (hd));
}

void
lcl_req_2args (lcl_conn_t *conn, uint32_t code, uint32_t callback,
	       uint32_t d1, uint32_t d2)
     /* Construct and send a protocol request with 2 arguments, namely
      * d1 and d2.  Those are passed in host order to this function
      * and it converts them to network order. */
{
  uint32_t hd[2];

  hd[0] = htonl (d1);
  hd[1] = htonl (d2);
  lcl_send_req (conn, code, callback, hd, sizeof (hd));
}

union lgd_proto_msg_u *
lcl_recv_rep (lcl_conn_t *conn)
     /* Receive a reply from the server and store it in a malloc()ated
      * request union. */
     /* XXX - Somehow I'm persuaded this function *must* be buggy. */
{
  uint32_t code, nlen;
  uint32_t hlen;
  union lgd_proto_msg_u *rep;

  /* XXX - I've never recisely understood what read() on a network
   * socket does in the absence of data: this assumes that it blocks
   * until the data has arrived.  If that's not the default behavior,
   * we're in *deep* trouble. */
  if ( read (conn->s, &code, sizeof (code)) < 0 )
    lcl_error ("Can't receive: %s\n", strerror (errno));
  if ( read (conn->s, &nlen, sizeof (nlen)) < 0 )
    lcl_error ("Can't receive: %s\n", strerror (errno));
  hlen = ntohl (nlen);
  if ( hlen < 8 )
    lcl_error ("Length too small\n");
  else if ( hlen > 4096 ) /* Protocol says: 1024 - we allow more */
    lcl_error ("Length too large\n");
  else if ( ( rep = malloc (hlen) ) == NULL )
    lcl_error ("Out of memory!\n");
  rep->f.code = code;
  rep->f.len = nlen;
  if ( read (conn->s, &rep->f.res, hlen-8) < 0 )
    lcl_error ("Can't receive: %s\n", strerror (errno));
  if ( hlen < 12 )
    rep->f.res = htonl (0);
  if ( hlen < 16 )
    rep->f.callback = 0;
  return rep;
}

union lgd_proto_msg_u *
lcl_recv_wanted_rep (lcl_conn_t *conn)
     /* Receive a ``wanted'' reply from the server.  The convention is
      * that replies with a callback of 0 are ``unwanted'' (they are
      * provoked by requests whose answer did not interest us), and we
      * ignore them.  Similarly, asynchronous events are unwanted and
      * we ignore them (following the protocol specification). */
{
  union lgd_proto_msg_u *rep;

  while ( 1 ) {
    rep = lcl_recv_rep (conn);
    if ( ( ntohl (rep->f.code) == LGD_PROTO_REP_ASY )
	 || ( rep->f.callback == 0 ) )
      free (rep);
    else
      return rep;
  }
}

int
lcl_recv_1dat (lcl_conn_t *conn, uint32_t *v)
     /* Receive a one-word response to an request.  Return 0 in case
      * of normal response, and the data is stored in v.  Return the
      * error code in case of error (-1 for generic error: this
      * shouldn't happen), and v is then untouched. */
{
  union lgd_proto_msg_u *rep;

  rep = lcl_recv_wanted_rep (conn);
  if ( ntohl (rep->f.code) == LGD_PROTO_REP_SYN )
    {
      if ( ntohl (rep->f.len) < 20 )
	*v = 0;
      else *v = ntohl (rep->f.data[0]);
      return 0;
    }
  else
    {
      if ( ntohl (rep->f.len) < 20 )
	return -1;
      else return ntohl (rep->f.data[0]);
    }
}

lcl_conn_t *
lcl_connect (char *server)
     /* Establish a connection with the named server.  This function
      * should be used prior to any other LCL function. */
{
  struct hostent *h;
  struct sockaddr_in name;
  union lgd_proto_msg_u *greet;
  lcl_conn_t *conn;

  conn = malloc (sizeof (lcl_conn_t));
  if ( ! conn )
    lcl_error ("Out of memory!\n");
  conn->server = strdup (server);
  h = gethostbyname (server);
  if ( ! h )
    lcl_error ("Can't connect to server.\n");
  if ( ( conn->s = socket (AF_INET, SOCK_STREAM, 0) ) < 0 )
    lcl_error ("Can't create socket: %s\n", strerror (errno));
  memset (&name, 0, sizeof (name));
  name.sin_family = AF_INET;
  name.sin_port = htons (PORTNUM);
  memcpy (&name.sin_addr, h->h_addr_list[0], h->h_length);
  if ( connect (conn->s, (struct sockaddr *) &name, sizeof (name)) < 0 )
    lcl_error ("Can't connect: %s\n", strerror (errno));
  greet = lcl_recv_rep (conn);
  if ( htonl (greet->f.code) != LGD_PROTO_REP_CON )
    lcl_error ("Greeting message has wrong code: %d\n",
	       htonl (greet->f.code));
  if ( htonl (greet->f.callback) != LGD_PROTO_MAGIC )
    lcl_error ("Greeting message has wrong magic number: %d\n",
	       htonl (greet->f.callback));
  free (greet);
  return conn;
}

void
lcl_req_nop (lcl_conn_t *conn)
     /* Send a NOP request and don't wait for an answer. */
{
  lcl_req_0arg (conn, LGD_PROTO_REQ_NOP, 0);
}

int
lcl_req_ack (lcl_conn_t *conn, uint32_t *v)
     /* Send an ACK? request, and wait for an answer if v is not NULL,
      * in which case put the answer in v. */
{
  if ( v ) /* We want an answer */
    {
      lcl_req_0arg (conn, LGD_PROTO_REQ_ACK, 1);
      return lcl_recv_1dat (conn, v);
    }
  else
    {
      lcl_req_0arg (conn, LGD_PROTO_REQ_ACK, 0);
      return 0;
    }
}
