/* This file is part of Legendes II */
/* Copyright (c) 1998-1999 by
 *   Antoine Allain (allain@aix.ensam.fr)
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File listener.c */

/* Thus speaketh ye mighty CVS:
 * $Id: listener.c,v 1.7 1999/05/30 22:23:03 david Exp $ */

/* This file contains the function that opens the initial socket and
 * the main function for the listener thread. */

/* CVS identity of this file. */
static char id[] = "$Id: listener.c,v 1.7 1999/05/30 22:23:03 david Exp $";

/* Include files */

#ifndef _THREAD_SAFE
#define _THREAD_SAFE		/* Make sure libc is reentrant. */
#endif

#include <stdio.h>		/* Standard I/O */
#include <stdlib.h>		/* Standard library functions */
#include <string.h>		/* For strerror() */

#include <errno.h>		/* Error numbers */

#include <sys/types.h>		/* System types */
#include <sys/socket.h>		/* Socket types */
#include <sys/time.h>		/* For struct timeval */
#include <unistd.h>		/* UNIX standard functions */
#include <fcntl.h>		/* For fcntl() and its constants */

#include <netinet/in.h>		/* INET domain (TCP/IP) sockets */

#include <netdb.h>		/* gethostbyaddr() and friends */

#include <pthread.h>		/* Posix threads */

#include <syslog.h>		/* For priority values */


#include "server.h"

/* Main functions */

void *
entry_listener (void *sck)
     /* Entry point for a listener thread */
{
  listener_loop (sck);
  return NULL;
}

void *
entry_handler (void *sck)
     /* Entry point for a handler thread */
{
  handler_loop (sck);
  return NULL;
}

void
init_gsock(void)
     /* Create the initial TCP/IP socket, bind() it to port PORTNUM,
      * and start listen()ing to it.  This function is exectuted in
      * single thread mode. */
{
  struct sockaddr_in name; /* To store socket ``name'' in */

  /* Create socket. */
  if ((gsock=socket(AF_INET,SOCK_STREAM,0))<0)
    {
      perror("Can't create socket()");
      exit(1);
    }
  /* Make it non-blocking.  See NOTE0006 for why we need this. */
  if ( fcntl (gsock, F_SETFL, O_NONBLOCK) < 0 )
    {
      perror("Can't make socket non-blocking");
      exit(1);
    }
  /* Bind it to all interfaces, TCP port PORTNUM. */
  memset(&name,0,sizeof(name));
  name.sin_family = AF_INET;
  name.sin_port = htons(PORTNUM); /* TCP port number for our server */
  name.sin_addr.s_addr = htonl (INADDR_ANY); /* Accept all interfaces */
  if (bind(gsock,(struct sockaddr *)&name,sizeof(name))<0)
    {
      perror("Can't bind() socket");
      exit(1);
    }
  /* Start listen()ing. */
  if (listen(gsock,10)<0)
    {
      perror("Can't listen() to socket");
      exit(1);
    }
}

void
accepter_loop (void)
     /* This function performs a loop that accepts incoming
      * connections and creates new threads for them. */
{
  fd_set fs;
  int max;
  int i;
  struct timeval tv;
  int s;
  struct sockaddr_in other;
  int other_namelen;
  struct hostent *other_host;

  while ( 1 )
    {
      FD_ZERO (&fs);
      FD_SET (gsock, &fs);
      max = gsock;
      /* Now do the waiting. */
      tv.tv_sec = 10; /* Timeout */
      tv.tv_usec = 0;
      if ( ( i = select (max+1, &fs, NULL, NULL, &tv) ) < 0 )
	{
	  if ( ( errno == EAGAIN ) || ( errno == EINTR ) )
	    continue;
	  /* What should we do?  It is hard to tell since we don't
	   * know what sort of error occurred.  We sleep(1) so as to
	   * avoid a busy-waiting loop in case the error doesn't
	   * vanish all by itself, and then restart. */
	  serverlog (LOG_ERR, "Error while select()ing: %s",
		     strerror (errno));
	  sleep (1);
	  continue;
	}
      if (i==0)
	/* Nothing happened (i.e. select() timed out).  We start
         * again. */
	{
	  sched_yield ();
	  continue;
	}
      /* Examine what we have found. */
      if ( FD_ISSET (gsock, &fs) )
	/* Global socket has something to say => new connection
         * possibly received */
	{
	  serverlog (LOG_INFO, "Accepting new connection");
	  other_namelen = sizeof(other);
	  if ( ( s = accept (gsock, &other, &other_namelen) ) < 0 )
	    {
	      if ( errno == EINTR )
		continue;
	      if ( errno == EAGAIN )
		{
		  serverlog (LOG_NOTICE, "This select() is worthless");
		  sched_yield ();
		  continue;
		}
	      serverlog (LOG_ERR, "Error while accept()ing: %s\n",
			 strerror (errno));
	      sched_yield ();
	      continue;
	    }
	  serverlog (LOG_DEBUG, "Port number: %d", other.sin_port);
	  /* XXX - gethostbyaddr() is not reentrant.  We should use
	   * gethostbyaddr_r() instead, shouldn't we?  Ugh... */
	  if ((other_host=gethostbyaddr((const char *)&other.sin_addr,
					sizeof(other.sin_addr),
					AF_INET))==NULL)
	    serverlog (LOG_INFO, "Can't resolve address");
	  else
	    serverlog (LOG_DEBUG, "Host name: %s", other_host->h_name);
	  /* Add this new connection to the list of connections. */
	  safe_pthread_mutex_lock (&lsock_mutex);
	  if (((lsock_tab = realloc(lsock_tab,
				    (++lsock_nb)*sizeof(struct lsock_s *)))
	       /* realloc() the lsock_tab to make space for a new
                * pointer */
	       ==NULL)||
	      ((lsock_tab[lsock_nb-1] = malloc(sizeof(struct lsock_s)))
	       /* And allocate something at the end of this new
                * pointer */
	       ==NULL))
	    {
	      perror("Out of memory");
	      exit(1);
	    }
	  /* Now initialize the data entries */
	  safe_pthread_mutex_init (&lsock_tab[lsock_nb-1]->d_mutex, NULL);
	  lsock_tab[lsock_nb-1]->sopen = 1;
	  lsock_tab[lsock_nb-1]->s = s;
	  lsock_tab[lsock_nb-1]->remote = other;
	  lsock_tab[lsock_nb-1]->handwork = 1;
	  pthread_cond_init (&lsock_tab[lsock_nb-1]->h_cond, NULL);
	  lsock_tab[lsock_nb-1]->state = CONN_STATE_NEW;
	  lsock_tab[lsock_nb-1]->buflen = 0;
	  lsock_tab[lsock_nb-1]->x = 0;
	  lsock_tab[lsock_nb-1]->y = 0;
	  pthread_create (&lsock_tab[lsock_nb-1]->listener, NULL,
			  entry_listener, lsock_tab[lsock_nb-1]);
	  pthread_create (&lsock_tab[lsock_nb-1]->handler, NULL,
			  entry_handler, lsock_tab[lsock_nb-1]);
	  /* Now release the lock.  !!! Do not do this earlier, for we
	   * use lsock_nb! */
	  safe_pthread_mutex_unlock(&lsock_mutex);
	}
      else
	/* Woa!  Vat iz ze meening of zis? */
	serverlog (LOG_WARNING, "This select() is really worthless!");
    }
}

void
listener_loop (struct lsock_s *sck)
     /* This function performs a loop that receives data from an
      * already established connection. */
{
  fd_set fs;
  int s;
  int max;
  int len;
  int i;
  struct timeval tv;

  while ( 1 )
    {
      if ( ! sck->sopen )
	/* Somebody has closed our socket.  Notice that we perform
	 * this test without holding the d_mutex, but that isn't
	 * important, because sopen can only go from 1 to 0, and if we
	 * miss the transition at this point, we will notice it later
	 * on. */
	{
	  serverlog (LOG_INFO, "Connection closed: dying.");
	  pthread_exit (NULL);
	}
      s = sck->s; /* Don't hold d_mutex: this doesn't move */
      max = s;
      FD_ZERO (&fs);
      FD_SET (s, &fs);
      /* Now do the waiting. */
      tv.tv_sec = 10; /* Timeout */
      tv.tv_usec = 0;
      if ( ( i = select (max+1, &fs, NULL, NULL, &tv) ) < 0 )
	{
	  if ( ( errno == EBADF ) || ( errno == EINTR )
	       || ( errno == EAGAIN ) )
	    {
	      serverlog (LOG_DEBUG, "Small error while select()ing: %s",
			 strerror (errno));
	      sched_yield ();
	      continue;
	    }
	  serverlog (LOG_ERR, "Error while select()ing: %s",
		     strerror (errno));
	  if ( safe_pthread_mutex_trylock (&sck->d_mutex) )
	    {
	      sched_yield ();
	      continue;
	    }
	  if ( sck->sopen )
	    {
	      close (s);
	      sck->sopen = 0;
	      sck->handwork = 1;
	      pthread_cond_signal (&sck->h_cond);
	    }
	  safe_pthread_mutex_unlock (&sck->d_mutex);
	  continue; /* Die next turn round */
	}
      if (i==0)
	/* Nothing happened (i.e. select() timed out.  We start
         * again. */
	{
	  sched_yield ();
	  continue;
	}
      /* Examine what we have found. */
      if ( FD_ISSET (s, &fs) )
	/* Data received on some socket */
	{
	  if (safe_pthread_mutex_trylock(&sck->d_mutex))
	    {
	      sched_yield();
	      continue;
	    }
	  if ( ! sck->sopen )
	    /* The connection has been closed.  We will die next time
             * round. */
	    {
	      safe_pthread_mutex_unlock (&sck->d_mutex);
	      sched_yield ();
	      continue;
	    }
	  len = BUFSIZE - sck->buflen; /* How much we try to read */
	  if ( len == 0 )
	    /* Buffer full.  We should give the handler threads some
	     * time to digest. */
	    {
	      safe_pthread_mutex_unlock(&sck->d_mutex);
	      serverlog (LOG_INFO, "Write buffer full.");
	      sched_yield();
	      continue;
	    }
	  if ( ( i = read (s, &(sck->buffer[sck->buflen]), len) ) < 0 )
	    {
	      close (s);
	      sck->sopen = 0;
	      sck->handwork = 1;
	      pthread_cond_signal (&sck->h_cond);
	      safe_pthread_mutex_unlock (&sck->d_mutex);
	      continue; /* Die next time round */
	    }
	  if ( i == 0 )
	    /* [NOTE0003] Memento: when a socket is shut down by the
	     * other party, select() always finds something
	     * interesting on that socket, but a read() returns 0.
	     * This is the only way to know a socket has shut down,
	     * apparently (.vomit.). */
	    {
	      serverlog (LOG_INFO, "Connection cleanly closed.");
	      shutdown (s, 2);
	      close (s);
	      sck->sopen = 0;
	      sck->handwork = 1;
	      pthread_cond_signal (&sck->h_cond);
	      safe_pthread_mutex_unlock (&sck->d_mutex);
	      continue; /* Die next time round */
	    }
	  /* At this point, we have really read data from the
           * socket. */
	  sck->buflen += i;
	  sck->handwork = 1;
	  pthread_cond_signal (&sck->h_cond);
	  safe_pthread_mutex_unlock(&sck->d_mutex);
	}
    }
}
